from typing import Any, Dict, List

from pydantic import BaseModel


class BaseElement(BaseModel):
    stype: str = ""
    sData: Dict[str, Any] = {}

class ElementManager:
    type_to_ele: Dict[str, BaseElement] = {}

    def register(
        self, 
        stype: str, 
        base: BaseElement
    ) -> None:
        if stype not in self.type_to_ele:
            self.type_to_ele[stype] = base

    def get(
        self, 
        stype: str
    ) -> BaseElement:
        return self.type_to_ele.get(stype, None)



from langchain.embeddings.openai import OpenAIEmbeddings

def Openai_embedding(conf):
    chatgpt_config = conf.get("models")["chatgpt"]
    openai_embedding = OpenAIEmbeddings(
        openai_api_key=chatgpt_config['token'],
    )
    return openai_embedding


from langchain.embeddings import HuggingFaceHubEmbeddings

def Huggingface_embedding(conf=None):
    repo_id = "sentence-transformers/all-MiniLM-L12-v2"
    huggingface_embedding = HuggingFaceHubEmbeddings(
        repo_id=repo_id,
        task="feature-extraction",
        huggingfacehub_api_token="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",
    )
    return huggingface_embedding


# class 
class Singleton(type):
    def __call__(cls, *args, **kwargs):
        if not hasattr(cls, '_instance'): 
            cls._instance = super().__call__(*args, **kwargs)
        return cls._instance

# wrapper 
def singleton(cls):
    instances = {}
    def wrapper(*args, **kwargs):
        if cls not in instances:
            instances[cls] = cls(*args, **kwargs)
        return instances[cls]
    return wrapper



"""Wrapper around HuggingFace Hub embedding models."""
from typing import Any, Dict, List, Optional

from pydantic import BaseModel

from crland.tools.embeddings.base import Embeddings

import requests
import json

class RemoteEmbeddings(BaseModel, Embeddings):
    """Wrapper around RemoteEmbeddings embedding models.   
    """
    url: str = "http://114.132.71.128:5008/predict" 
    # all-MiniLM-L12-v2
    """Model name to use."""
    task: Optional[str] = "semantic"
    """Task to call the model with."""
    model_kwargs: Optional[dict] = None
    """Key word arguments to pass to the model."""

    def prepare_input(self, text) -> str:
        data_json = '{"header":{"req_id":"123"},"texts":"' + text + '"}'
        return data_json

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        """Call out to RemoteEmbeddings's embedding endpoint for embedding search docs.

        Args:
            texts: The list of texts to embed.

        Returns:
            List of embeddings, one for each text.
        """
        responses = []
        for text in texts:
            responses.append(self.embed_query(text))

        return responses

    def embed_query(self, text: str) -> List[float]:
        """Call out to RemoteEmbeddings's embedding endpoint for embedding query text.

        Args:
            text: The text to embed.

        Returns:
            Embeddings for the text.
        """
        response = requests.post(url=self.url, 
                                  data=self.prepare_input(text).encode('utf-8'), 
                                  timeout=3)

        return json.loads(response.text)["embedding"]
        """
        return [1.0, 1.0, 1.0]
        """


# -*- coding: utf-8 -*-
# (C) Run, Inc. 2022
# All rights reserved (Author BinZHU)
# Licensed under Simplified BSD License (see LICENSE)

import os
import sys
import logging
from logging.handlers import RotatingFileHandler

def get_logger (filename, 
                rotating=True, 
                verbosity=1, 
                name=None):
    r'''Define C type logging method`.
    
    :param filename: filename where logs are stored.
    :param verbosity: whether verbosity infos are required.
    :param name: name of logger
    :return: logger instance with usage logger.info(msg)
    '''
    level_dcit = {
        0: logging.DEBUG,
        1: logging.INFO,
        2: logging.WARNING
    }
    formatter = logging.Formatter(
        "[%(asctime)s][%(filename)s][line:%(lineno)d][%(levelname)s] %(message)s"
    )
    logger = logging.getLogger(name)
    logger.setLevel(level_dcit[verbosity])
    
    save_path = os.path.join(os.path.dirname(__file__), '../logs/')
    file_handle = logging.FileHandler(save_path + filename, "w")
    if rotating:
        file_handle = RotatingFileHandler (filename=save_path + filename, 
                                           maxBytes=1024*1024*100, 
                                           backupCount=10)  
    file_handle.setFormatter(formatter)
    logger.addHandler(file_handle)
    
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)
    logger.addHandler(stream_handler)

    return logger


from langchain import Anthropic, OpenAI
from langchain.chat_models import ChatOpenAI

from crland.models.guanaco import Guanaco

class ModelManager:
    def __init__(self, config):
        self.config = config
        #print(config.config)
        self.models = {}
        if "chatgpt" in  self.config.get("models"):
            chatgpt_config = self.config.get("models")["chatgpt"]
            self.register_model(
                "chatgpt", 
                ChatOpenAI(
                    temperature = 0, 
                    openai_api_key = chatgpt_config["token"], 
                    model_name = chatgpt_config["model_name"],
                    openai_api_base = chatgpt_config["api_base"],
                    max_tokens = chatgpt_config["max_tokens"] if "max_tokens" in chatgpt_config else 1024,
                    )
                )
        if "vicuna" in  self.config.get("models"):
            self.register_model(
                "vicuna", 
                OpenAI(
                    openai_api_key=self.config.get("models")["vicuna"]["token"], 
                    model_name="vicuna-13b",
                    openai_api_base="http://10.8.14.49:8007/v1",
                    max_tokens_to_sample = 1024,
                    temperature=0
                    )
                )
        if "qwen" in  self.config.get("models"):
            self.register_model(
                "qwen", 
                OpenAI(
                    openai_api_key=self.config.get("models")["qwen"]["token"], 
                    model_name="Qwen",
                    openai_api_base="http://10.8.14.49:7861/v1",
                    max_tokens_to_sample = 1024,
                    temperature=0
                    )
                )
        
        if "guanaco" in self.config.get("models"):
            self.register_model(
                "guanaco", Guanaco()
            )
            
    def register_model(self, model_name, model_class):
        self.models[model_name] = model_class
        
        if not self.config.get(model_name):
            self.config.set(model_name, {})
            
    def get_model(self, model_name):
        return self.models[model_name]




import random
import json
from typing import Any, Dict, List
from crland.searchhub.page_manage.base import BaseElement
from crland.datacenter.knowledge.graph import Graph
from crland.datacenter.knowledge.executor import Executor

class Model(BaseElement):
    stype: str = "Model"

class ModelManager:
    type_to_ele: Dict[str, Model] = {}
    graph: Graph = None

    def __init__(
        self,
        graph
    ):
        self.graph = graph

    def register(
        self, 
        name: str, 
        base: Model
    ) -> None:
        if name not in self.type_to_ele:
            self.type_to_ele[name] = base

    def get(
        self, 
        name: str
    ) -> Model:
        return self.type_to_ele.get(name, None)

    # will be moved to db later
    def get_details(
        self, 
        data: Dict[str, Any]
    ) -> Dict[str, Any]:
        #print(data)
        return {
            "model": data["model"],
            "icon": "",
            "author": data["signature"],
            "tag": data['tag'],
            "time": "Updated " + str(random.randint(1,30)) + " days ago",
            "download": str(random.randint(1,1000)),
            "like": str(random.randint(1,200)),
            "jump_url": "",
        }

    def get_homepage(
        self, 
        num: int
    ) -> Dict[str, Any]:
        result = []
        for (name, exe) in self.graph.get_all_exec():
            result.append(self.get_details({
                "model": name,
                "tag": exe.description,
                "signature": exe.signature
            }))
        return result


    def get_factor_model(
        self, 
        factor: str, 
        num: int
    ) -> Dict[str, Any]:
        result = []
        for (name, exe) in self.graph.get_factor_exec(factor):
            result.append(self.get_details({
                "model": name,
                "tag": exe.description,
                "signature": exe.signature
            }))
        return result


#print(ModelManager().get_homepage(20))

#data = ModelManager().get_factor_model("Company Analysis", 20)



from typing import Any, Dict, List, Optional, Callable
from crland.datacenter.knowledge.executor import Executor

class Factor:
    '''
        class for factor, used in graph
    '''    
    def __init__(
        self,
        name: str,
        description: str,
        recall: List[str] = [],
        parents: List['Factor'] = [],
        childrens: List['Factor'] = [],
        executors: List[Executor] = []
    ):
        self.name = name
        self.description = description
        self.recall = recall
        self.parents = parents
        self.childrens = childrens
        self.executors = executors

    @classmethod
    def create(
        cls,
        name: str,
        description: str
    ) -> 'Factor':
        return cls(name=name, description=description)
    
    def register_func(
        self, 
        func: Executor
    ):
        self.executors.append(func)
    
    def add_parents(
        self, 
        parent: 'Factor'
    ):
        if parent not in self.parents:
            self.parents.append(parent)

    def get_parents(
        self
    ) -> List['Factor']:
        return self.parents

    def add_childrens(
        self, 
        child: 'Factor'
    ):
        if child not in self.childrens:
            self.childrens.append(child)

    def get_childrens(
        self
    ) -> List['Factor']:
        return self.childrens    


"""Implement an LLM driven browser."""
from __future__ import annotations

import warnings
import copy
from typing import Any, Dict, List, Optional

from pydantic import Extra
from pydantic import BaseModel, root_validator

from langchain.base_language import BaseLanguageModel
from langchain.callbacks.manager import CallbackManagerForChainRun
from langchain.chains.base import Chain

from crland.agent.recall.base import FuncRecallChain
from crland.config import Config
from crland.models.base import ModelManager

class CollectChain(Chain):

    func_chain: FuncRecallChain
    input_key: str = "input"
    output_key: str = "result"
    query_key: str = "query"
    """Objective that NatBot is tasked with completing."""
    """[Deprecated] LLM wrapper to use."""

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.forbid
        arbitrary_types_allowed = True

    @classmethod
    def from_chains(
        cls, llm: BaseLanguageModel, manager: ModelManager, **kwargs: Any
    ) -> CollectChain:
        """Load from LLM."""
        func_chain = FuncRecallChain.from_llm_index(llm=llm, index_manager=manager)
        return cls(func_chain=func_chain, **kwargs)

    @property
    def input_keys(self) -> List[str]:
        """Expect url and browser content.

        :meta private:
        """
        return [self.input_key, self.query_key]

    @property
    def output_keys(self) -> List[str]:
        """Return command.

        :meta private:
        """
        return [self.output_key]

    def _call(
        self,
        inputs: Dict[str, str],
        run_manager: Optional[CallbackManagerForChainRun] = None,
    ) -> Dict[str, str]:
        _run_manager = run_manager or CallbackManagerForChainRun.get_noop_manager()

        require_data = inputs[self.input_key]
        query = inputs[self.query_key]
        query_data = f"""Question: {query}\n```\n"""
        for k, v in require_data.items():
            func_inputs = copy.deepcopy(inputs)
            func_inputs.update({"input": k})
            k_ret = self.func_chain(func_inputs)
            print(k, k_ret)
            func_ret = k_ret['func']
            if isinstance(func_ret, dict):
                func_ret = func_ret['result'] # for function return with chart
            query_data += k + "\n" + func_ret + "\n\n"

        query_data += """```\nTry you best to answer reasonable and helpfully in Chinese, think step by step."""
        return {self.output_key: query_data}

    @property
    def _chain_type(self) -> str:
        return "collect_chain"


from typing import Dict, Any, Union, List

def stack_line(inputs: Union[Dict[str, Any], List[Any]], 
        labels: Dict[str, str]) -> Dict[str, Any]:
    """ translate dataformat to echart format for bar
        :param inputs: Data from database（x, y）
        :param labels: Identify labels of Data (title)
        :return option: echart format
    """
    x = []
    y = {}
    legend = labels["y"]
    if isinstance(legend, str):
        legend = [legend]
    if isinstance(inputs, List):
        xlabel = labels['x']
        for i in inputs:
            x.append(i[xlabel])
            for j in legend:
                if j not in y:
                    y[j] = []
                y[j].append(i[j])
    series = [
        {
            "name": x,
            "type": 'line',
            #"stack": 'Total',
            "data": y[x]
        } for x in legend
    ]
    option = {
        "title": {
            "text": labels["title"],
        },
        "tooltip": {
            "trigger": 'axis'
        },
        "legend": {
            "data": legend
        },
        "grid": {
            "left": '3%',
            "right": '4%',
            "bottom": '3%',
            "containLabel": True
        },
        "xAxis": {
            "type": 'category',
            "boundaryGap": False,
            "data": x
        },
        "yAxis": {
            "type": 'value'
        },
        "series": series
    }
    return option


# flake8: noqa
PREFIX = """Answer the following questions. You have access to the following functions:"""
FORMAT_INSTRUCTIONS = """Use the following format:

Question: the input question you must answer
Thought: you should think which function you should use and only choose one
Function: the function to choose, should be one of [{tool_names}]
Function Input: the input parameters to the function is an entity extracted from """

SUFFIX = """
You should only find one. Begin!

Question: {input}
Thought:{agent_scratchpad}"""


import re
from typing import Union

from langchain.agents.agent import AgentOutputParser
from crland.agent.rank.prompt import FORMAT_INSTRUCTIONS
from langchain.schema import AgentAction, AgentFinish, OutputParserException

class FunctionOutParser(AgentOutputParser):
    def get_format_instructions(self) -> str:
        return FORMAT_INSTRUCTIONS

    def parse(self, text: str) -> Union[AgentAction, AgentFinish]:
        print(text)
        # \s matches against tab/newline/whitespace
        regex = (
            r"Function\s*\d*\s*:[\s]*(.*?)[\s]*Function\s*\d*\s*Input\s*\d*\s*:[\s]*(.*)"
        )
        match = re.search(regex, text, re.DOTALL)
        if not match:
            raise OutputParserException(f"Could not parse LLM output: `{text}`")
        action = match.group(1).strip().replace("\\", "")
        action_input = re.sub(r'\([^)]*\)', '', match.group(2)).replace('"', "").strip(" ") # delete parathese and its contents
        #print(action, "\n", action_input)
        return AgentAction(action, action_input, text)


from typing import Dict, Any

from langchain.callbacks.base import BaseCallbackHandler
from langchain.schema import (
    AgentAction,
    AgentFinish,
)
import asyncio

from crland.service.error import StatusCodeEnum
from crland.service.base import wrapper_return

class WebsocketCallbackHandler(BaseCallbackHandler):
    def __init__(self, socket, loop):
        self.socket = socket
        self.loop = loop

    def on_chain_end(self, outputs: Dict[str, Any], **kwargs: Any) -> None:
        """Run when chain ends running."""
        print("************* output *************\n", outputs)
        self.loop.create_task(
            self.socket.send(
            wrapper_return(outputs["text"], StatusCodeEnum.OK)
        ))

    """
    def on_agent_finish(self, finish: AgentFinish, **kwargs: Any) -> None:
        self.loop.create_task(
            self.socket.send(
            wrapper_return(outputs["text"], StatusCodeEnum.OK)
        ))
    """


#!/usr/bin/env python
# -*- coding: utf-8 -*-
import pandas as pd
from sqlalchemy import create_engine
from sqlalchemy import text
from sqlalchemy.types import NVARCHAR, Float, Integer
import pymysql.cursors

from crland.config import Config
from crland.datacenter.database import EMPTY_DATA, EMPTY_NUM

_bound_num = 10000

class Database:
    """ Python连接到 MySQL 数据库及相关操作 """
    
    connected = False
    __conn = None
    engine = None

    # 构造函数，初始化时直接连接数据库
    def __init__(self, conf):
        if type(conf) is not dict:
            print('错误: 参数不是字典类型！')
        else:
            for key in ['host', 'port', 'user', 'pw', 'db']:
                if key not in conf.keys():
                    print('错误: 参数字典缺少 %s' % key)
            if 'charset' not in conf.keys():
                conf['charset'] = 'utf8'
        try:
            self.__conn = pymysql.connect(
                host=conf['host'],
                port=conf['port'],
                user=conf['user'],
                passwd=conf['pw'],
                db=conf['db'],
                charset=conf['charset'],
                cursorclass=pymysql.cursors.DictCursor)
            self.connected = True
            self.engine = create_engine(
                'mysql+pymysql://%(user)s:%(pw)s@%(host)s:%(port)d/%(db)s?charset=utf8' % conf
            )         
        except pymysql.Error as e:
            print('数据库连接失败:', end='')

    def create_table(self, table, contents):
        try:
            #使用execute()方法执行sql，如果表存在则删除
            with self.__conn.cursor() as cursor:             
                cursor.execute('drop table if EXISTS ' + table)
            #创建表的sql
                if isinstance(contents, dict):
                    new_contents = ""
                    for k, v in contents.items():
                        new_contents += k + " " + v + ","
                    contents = new_contents[:-1] #delete last ","
                sql = '''
                    create table ''' + table + '''(
                    ''' + contents + '''
                    )
                '''
                print(sql)
                cursor.execute(sql)
        except pymysql.Error as e:
            self.__conn.rollback()
            return False

    def add_column_to_table(self, table, contents):
        try:
            with self.__conn.cursor() as cursor:
                if isinstance(contents, dict):
                    new_contents = ""
                    for k, v in contents.items():
                        new_contents += k + " " + v + ","
                    contents = new_contents[:-1] #delete last ","
                sql = '''ALTER TABLE ''' + table + ''' ADD ''' + contents
                print(sql)
                cursor.execute(sql)
        except pymysql.Error as e:
            self.__conn.rollback()
            return False

    # 插入数据到数据表
    def insert(self, table, val_obj):
        sql_top = 'INSERT INTO ' + table + ' ('
        sql_tail = ') VALUES ('
        try:
            for key, val in val_obj.items():
                sql_top += key + ','
                if isinstance(val, str):
                    val = "\"" + val + "\"" 
                sql_tail += val + ','
            sql = sql_top[:-1] + sql_tail[:-1] + ')'
            #print(sql)
            with self.__conn.cursor() as cursor:
                cursor.execute(sql)
            self.__conn.commit()
            return self.__conn.insert_id()
        except pymysql.Error as e:
            self.__conn.rollback()
            return False

    # 插入数据到数据表
    def execute(self, sql):
        try:
            with self.__conn.cursor() as cursor:
                cursor.execute(sql)
            self.__conn.commit()
            return True
        except pymysql.Error as e:
            self.__conn.rollback()
            return False

    # 更新数据到数据表
    def update(self, table, val_obj, range_str):
        sql = 'UPDATE ' + table + ' SET '
        try:
            for key, val in val_obj.items():
                sql += key + '=' + val + ','
            sql = sql[:-1] + ' WHERE ' + range_str
            with self.__conn.cursor() as cursor:
                cursor.execute(sql)
            self.__conn.commit()
            return cursor.rowcount
        except pymysql.Error as e:
            self.__conn.rollback()
            return False

    # 删除数据在数据表中
    def delete(self, table, range_str):
        sql = 'DELETE FROM ' + table + ' WHERE ' + range_str
        try:
            with self.__conn.cursor() as cursor:
                cursor.execute(sql)
            self.__conn.commit()
            return cursor.rowcount
        except pymysql.Error as e:
            self.__conn.rollback()
            return False

    # 查询唯一数据在数据表中
    def select_one(self, table, factor_str, field='*'):
        sql = 'SELECT ' + field + ' FROM ' + table + ' WHERE ' + factor_str
        try:
            with self.__conn.cursor() as cursor:
                cursor.execute(sql)
            self.__conn.commit()
            return cursor.fetchall()[0]
        except pymysql.Error as e:
            return False

    # 查询多条数据在数据表中
    def select_more(self, table, range_str, field='*'):
        sql = 'SELECT ' + field + ' FROM ' + table + ' WHERE ' + range_str
        try:
            with self.__conn.cursor() as cursor:
                cursor.execute(sql)
            self.__conn.commit()
            return cursor.fetchall()
        except pymysql.Error as e:
            return False

    # 查询条件筛选查询
    def select_by_order(self, table, order_column, limit_num=4, field='*'):
        inner_sql = 'SELECT ' + field + ' FROM ' + table + \
                    ' order by ' + order_column + ' desc limit ' + str(limit_num)
        sql = "SELECT * from (" + inner_sql + ")t order by " + order_column
        #print(sql)
        try:
            with self.__conn.cursor() as cursor:
                cursor.execute(sql)
            self.__conn.commit()
            return cursor.fetchall()
        except pymysql.Error as e:
            print(e)
            return False

    # 直接执行
    def exec(self, sql):
        try:
            with self.__conn.cursor() as cursor:
                cursor.execute(sql)
            self.__conn.commit()
            return cursor.fetchall()
        except pymysql.Error as e:
            print(e)
            return False

    # 统计某表某条件下的总行数
    def count(self, table, range_str='1'):
        sql = 'SELECT count(*)res FROM ' + table + ' WHERE ' + range_str
        try:
            with self.__conn.cursor() as cursor:
                cursor.execute(sql)
            self.__conn.commit()
            return cursor.fetchall()[0]['res']
        except pymysql.Error as e:
            return False

    # 统计某字段（或字段计算公式）的合计值
    def sum(self, table, field, range_str='1'):
        sql = 'SELECT SUM(' + field + ') AS res FROM ' + table + ' WHERE ' + range_str
        try:
            with self.__conn.cursor() as cursor:
                cursor.execute(sql)
            self.__conn.commit()
            return cursor.fetchall()[0]['res']
        except pymysql.Error as e:
            return False

    # 销毁对象时关闭数据库连接
    def __del__(self):
        try:
            self.__conn.close()
        except pymysql.Error as e:
            pass

    # 关闭数据库连接
    def close(self):
        self.__del__()


    def insert_data_by_pandas(
        self, 
        dataframe, 
        table_name, 
        dtypes={}, 
        if_exists='append', 
        single=False
    ):
        '''
        通过dataframe 向 sql 中插入表，此方法缺点是若表已存在，不能替换表中部分重复数据，只能替换/追加整张表
        :param dataframe: pd.Dataframe类型
        :param table_name: 插入的表名
        :param if_exists: {'fail', 'replace', 'append'}, default 'fail'
            - fail: If table exists, do nothing.
            - replace: If table exists, drop it, recreate it, and insert data.
            - append: If table exists, insert data. Create if does not exist.
        :return:
        '''

        def create_upsert_method(primary_key_column):
            def method(table, conn, keys, data_iter):
                # Create the INSERT statement with named placeholders
                insert = f"INSERT INTO {table.name} ({','.join(keys)})"
                try:
                    for d in data_iter:
                        values = " VALUES("
                        update = " ON DUPLICATE KEY UPDATE "                        
                        for i, name in enumerate(keys):
                            new_val = "NULL,"
                            if d[i]:
                                new_val = "'" + str(d[i]) + "',"
                            values += new_val
                            if name not in primary_key_column:
                                update += name + "=" + new_val
                        
                        values = values[:-1] + ")"
                        sql = insert + values + update[:-1]
                        conn.execute(text(sql))
                except Exception as e:
                    print("An error occurred:", str(e))
            return method

        if single:
            for i in range(len(dataframe)):
                try:
                    dataframe.iloc[i:i+1].to_sql(
                        table_name, 
                        self.engine, 
                        if_exists=if_exists, 
                        index=False, 
                        chunksize=100,
                        dtype = dtypes,
                        method = create_upsert_method(list(dtypes.keys()))
                    )
                except:
                    print("e")
                    pass
        else:
            try:
                dataframe.to_sql(
                    table_name, 
                    self.engine, 
                    if_exists=if_exists, 
                    index=False, 
                    chunksize=100,
                    dtype = dtypes,
                    method = create_upsert_method(list(dtypes.keys()))                    
                )
            except:
                pass 

    # 查询多条数据在数据表中
    def get_data_and_manuel_summary(
        self, 
        table, 
        order_str, 
        limit_num=3, 
        columns_to_names = None, 
        with_chart=False
    ):
        fields = ""
        if columns_to_names is None:
            fields = "*"
        else:
            fields = ", ".join(k for k, v in columns_to_names.items())
        # order name not in show
        if order_str not in fields:
            fields += ", " + order_str
        data = self.select_by_order(
            table = table, 
            order_column = order_str,
            limit_num=limit_num,            
            field = fields
            )
        #print(data, columns_to_names)
        idx = 0         
        for line in data[-3:]:
            for k, v in line.items():
                if k in columns_to_names:
                    if 0 == idx:
                        columns_to_names[k] = columns_to_names[k] + " in temporal order:"
                    try:
                        if abs(v) > _bound_num:
                            columns_to_names[k] += "{:.1e}".format(v) + ", " # to short it number                            
                        else:
                            columns_to_names[k] += str(round(float(v), 1)) + ", " # to short it number
                    except:
                        columns_to_names[k] += str(v) + ", "
            idx += 1
        result = ""
        if fields == "*":
            result = " \n".join("The " + k + " : " + v for k, v in columns_to_names.items())
            # error here, to improve later
        else:
            result = " \n".join(v[:-2] for k, v in columns_to_names.items())
        if with_chart:
            return {
                "result": result,
                "chart": data
            }
        else:
            return result

    # 直接执行sql查询多条数据在数据表中
    def exec_by_modelformat(
        self, 
        sql, 
        columns_to_names
    ):
        data = self.exec(sql)
        if not data:
            return EMPTY_DATA
        # print(data, columns_to_names)
        hit_columns = dict()
        idx = 0         
        for line in data:
            for k, v in line.items():
                if k in columns_to_names:
                    if 0 == idx:
                        hit_columns[k] = "Progressive " + columns_to_names[k] + ":"
                    try:
                        if abs(v) > _bound_num:
                            hit_columns[k] += "{:.1e}".format(v) + ", " # to short it number                            
                        else:
                            hit_columns[k] += str(round(float(v), 1)) + ", " # to short it number
                    except:
                        hit_columns[k] += str(v) + ", "
            idx += 1
        result = " \n".join(v[:-2] for k, v in hit_columns.items())
        return result


import json
from typing import Any, Dict, List
from crland.searchhub.page_manage.base import BaseElement
from crland.datacenter.knowledge.graph import Graph
class Factor(BaseElement):
    stype: str = "Factor"

class FactorManager:
    type_to_ele: Dict[str, Factor] = {}
    graph: Graph = None

    def __init__(
        self, 
        graph
    ):
        self.graph = graph

    def register(
        self, 
        name: str, 
        base: Factor
    ) -> None:
        if name not in self.type_to_ele:
            self.type_to_ele[name] = base

    def get(
        self, 
        name: str
    ) -> Factor:
        return self.type_to_ele.get(name, None)

    # will be moved to db later
    def get_details(
        self, 
        factor: str
    ) -> Dict[str, Any]:
        return {
            "factor": factor,
            "icon": "",
            "jump_url": "",
        }

    def get_homepage(
        self, 
        num: int
    ) -> Dict[str, Any]:
        '''
            level 2 category
        '''
        result = {}
        for head in self.graph.headers:
            result[head] = []
            for ele in self.graph.get_factor(head).get_childrens():
                result[head].append(
                    self.get_details(
                        ele.name
                    )
                )
        return result


from pydantic import BaseModel
from typing import  Callable, Any, Dict, List, Union, Tuple
from crland.searchhub.task import candidate_tasks
from crland.searchhub.task.base import Task

class TaskManager():
    tasks: Dict[str, Task] = {}
    def __init__(
        self,
        config
    ):
        self.config = config
        tasks = self.config.get('task')
        for task in candidate_tasks:
            if task.name in tasks:
                self.tasks[task.name] = task()
    
    def get_tasks(
        self
    ) -> List[str]:
        '''
            default is for non task chat
        '''
        return list(self.tasks.keys()) + ["default"]

    def get_task_by_name(
        self,
        name: str
    ) -> Task:
        return self.tasks.get(name, None)

    def extract_task(
        self,
        query: str
    ) -> Dict[str, str]:
        """
        Extract task name and fix query for input
        """
        for t in self.tasks.keys():
            if query.startswith("@" + t):
                return {
                    "task": t,
                    "query": query[len(t)+1:]
                }
        return {"query": query}


from typing import Dict
from crland.config import Config
from crland.datacenter.knowledge.graph import Graph
from crland.datacenter.knowledge.executor import ExecutorManager
from crland.searchhub.recall.base import IndexManager
from crland.datacenter.knowledge.executor import ExecutorManager
from crland.searchhub.recall.channel.faiss import FAISS2
from crland.models.base import ModelManager
from crland.tools.embeddings.embedding_manager import EmbeddingManager
from crland.agent.recall.base import FuncRecallChain
from crland.agent.collect.base import CollectChain
from crland.agent.base import JustModel
#from crland.agent.analysis.base import AnalysisChain
from crland.searchhub.task.base import Task

class AnalysisTask(Task):
    name = "analysis"
    def __init__(
        self
    ):
        self.config = Config()
        self.model_manager = ModelManager(config=self.config)
        self.llm = self.model_manager.get_model("chatgpt")
        #self.analysis_chain = AnalysisChain.from_llm(self.llm)
        self.analysis_chain = None
        self.graph = Graph.build_from_file(
            "crland/datacenter/knowledge/schema.md"
        )
        self.graph.assemble(ExecutorManager())
        db = FAISS2.from_texts(
            inputs = self.graph.get_available_executors(), 
            embedding = EmbeddingManager.get_embedding(self.config)
        )
        index_manager = IndexManager()
        index_manager.register(self.config.get('index'), db)
        #db.save_local("./crland/datacenter/local_data")
        #db = FAISS2.load_local("./crland/datacenter/local_data", embedding)
        self.collect_chain = CollectChain.from_chains(
            self.llm, 
            index_manager
        )

    def execute(
        self, 
        text, 
        **kwargs
    ) -> Dict[str, str]:
        #require_data = self.analysis_chain(
        #    text
        #)["text"]
        require_data = ExecutorManager().name_to_executor
        collect_data = self.collect_chain({
            "input": require_data,
            "channel": self.config.get('index')["channel"],
            "query": text,
            "params": kwargs.get('company', "")
        })["result"]
        print(collect_data)
        JustModel(self.llm)(collect_data)['result']
        return {
            "intermediate": collect_data,
            "result": result
        }

if __name__ == '__main__':
    task = AnalysisTask()
    result = task.execute("", company = "xx")
    print(result)
    #graph = Graph.build_from_file("crland/datacenter/knowledge/schema.md")
    #graph.assemble(analysis_func)
    #print(graph.get_available_func()) 




from typing import Any, Dict, List, Optional, Callable
from crland.tools.singleton import singleton

class Executor:
    '''
        class for factor functions, support user defined
    '''
    def __init__(
        self,
        name: str,
        signature: str,
        description: str,        
        func: Callable,
        graph_node: bool = True,
        stats: Dict = {}
    ):
        self.name = name
        self.signature = signature
        self.func = func
        self.description = description
        self.graph_node = graph_node
        self.stats = stats

    @classmethod
    def create(
        cls,
        name,
        signature,
        description,
        func,
        graph_node,
    ) -> 'Executor':
        return cls(
            name=name, 
            signature=signature, 
            description=description, 
            func=func,
            graph_node=graph_node
            )
    
    @classmethod
    def create_from_func(
        cls,
        func,
        description,
    ) -> 'Executor':
        return cls(
            name=func.__name__, 
            signature="default", 
            description = description, 
            func=func
            )

@singleton
class ExecutorManager:
    name_to_executor = {}
    
    def register(
        self, 
        name,
        signature,
        exe
    ):
        self.name_to_executor[name + "|" + signature] = exe

    def register(
        self, 
        name,
        func,
        description,
        signature,
        graph_node        
    ):
        key = name + "|" + signature
        if key in self.name_to_executor:
            raise f"Key already occurs: {key}"

        self.name_to_executor[name + "|" + signature] = Executor.create(
            name,
            signature,
            description,
            func,
            graph_node
        )
    
    def get(
        self, 
        name, 
        signature="default"
    ):
        return self.name_to_executor.get(name + "|" + signature, None)

    def build_recall(
        self
    ):
        recall_to_exe = {}
        for name, exe in self.name_to_executor.items():
            if exe.name:
                recall_to_exe[exe.name] = exe
            if exe.description:
                recall_to_exe[exe.description] = exe
        return recall_to_exe


from typing import Dict
from crland.config import Config
from crland.models.base import ModelManager
from crland.searchhub.recall.base import IndexManager
from crland.searchhub.recall.channel.faiss import FAISS2
from crland.datacenter.database.data_read.macro import *
from crland.datacenter.knowledge.executor import ExecutorManager
from crland.tools.embeddings.embedding_manager import EmbeddingManager
from crland.agent.recall.base import FuncRecallChain
from crland.searchhub.task.base import Task

class SearchTask(Task):
    name = "search"
    def __init__(
        self
    ):
        self.config = Config()
        self.model_manager = ModelManager(config=self.config)
        self.llm = self.model_manager.get_model("guanaco")
        index_manager = IndexManager()
        #print(ExecutorManager().build_recall())
        #db = FAISS2.from_texts (
        #    inputs = ExecutorManager().build_recall(),
        #    embedding = EmbeddingManager.get_embedding(self.config)
        #)
        #index_manager.register(self.config.get('index'), db)
        #db.save_local("./crland/datacenter/local_data")
        #db = FAISS2.load_local("./crland/datacenter/local_data", embedding)
        self.chain = FuncRecallChain.from_llm_index(
            self.llm, 
            index_manager
        )

    def execute(
        self, 
        text, 
        **kwargs
    ) -> Dict[str, str]:
        result =  self.chain({
            "input": text, 
            "channel": self.config.get('index')["channel"], 
            "query": text,
            "params": kwargs.get('company', "深圳万象城")
            })['func']
        if isinstance(result, dict):
            return result
        else:
            return {
                "result": result
            }

if __name__ == '__main__':
    task = SearchTask()
    result = task.execute("查询深圳万象城的运营情况")
    print(result)


import random
import json
from typing import Any, Dict, List
from crland.searchhub.page_manage.base import BaseElement
from crland.tools.redis.redis_tools import RedisManager


class User(BaseElement):
    stype: str = "User"

class UserManager:
    def __init__(
        self, 
        config
    ):
        self.type_to_ele: Dict[str, User] = {}
        self.redis_manager = RedisManager(config=config)

    def register(
        self, 
        name: str, 
        base: User
    ) -> None:
        if name not in self.type_to_ele:
            self.type_to_ele[name] = base

    def get(
        self, 
        name: str
    ) -> User:
        return self.type_to_ele.get(name, None)

    def get_history(
        self, 
        session_id: str
    ) -> Dict[str, Any]:
        """
            For chat front message
        """
        result = []
        for d in self.redis_manager.get("session").lrange(session_id, 0, -1):
            result.append(
                {
                    "input": d["input"],
                    "output": d["output"]
                }
            )
        return result

    def get(
        self, 
        name: str
    ) -> User:
        return self.type_to_ele.get(name, None)

    def get_llm_history(
        self, 
        session_id: str,
        limit = 2
    ) -> Dict[str, Any]:
        """
            For chat front message
        """
        result = ""
        for d in self.redis_manager.get("session").lrange(session_id, -1-limit, -1):
            d = json.loads(d)
            result += "Human: " + d.get("intermediate", d["input"]) + "\n"
            result += "Assitant: " + d.get("output", "") + "\n"
        return result + "Human: "

    def insert_history(
        self, 
        session_id, 
        json_data
    ):
        #print(session_id, json_data)        
        self.redis_manager.get("session").rpush(session_id, json_data)  

    def get_snapshot(
        self, 
        user: str
    ) -> Dict[str, Any]:
        result = []
        try:
            return self.redis_manager.get("chatshot").lrange(user, 0, -1).decode('utf-8')
        except:
            return result

    def insert_snapshot(
        self, 
        user: str, 
        session_id: str, 
        query: str
    ):
        #print(user, session_id, query)
        if not self.redis_manager.get("session").exists(session_id):
            self.redis_manager.get("chatshot").rpush(user, json.dumps({
                "session_id": session_id,
                "desc": query
            }))


from typing import Dict, Any, Union, List

def bar(inputs: Union[Dict[str, Any], List[Any]], 
        labels: Dict[str, str]) -> Dict[str, Any]:
    """ translate dataformat to echart format for bar
        :param inputs: Data from database
        :param labels: Identify labels of Data
        :return option: echart format
    """
    x = []
    y = []
    if isinstance(inputs, List):
        xlabel = labels['x']
        ylabel = labels['y']
        for i in inputs:
            x.append(i[xlabel])
            y.append(i[ylabel])
    option = {
        "title": {
          "text": labels["title"],
          "left": "center"
        },
        "xAxis": {
            "type": 'category',
            "name": labels['x'],
            "data":  x
        },
        "yAxis": {
            "type": 'value',
            "name": labels['y']
        },
        "series": [
            {
                "data": y,
                "type": 'bar',
                "showBackground": True,
                "backgroundStyle": {
                    "color": 'rgba(180, 180, 180, 0.2)'
                }
            }
        ]
    }
    return option


from typing import Any, Dict, List, Optional

import requests
from pydantic import BaseModel, Extra, root_validator

from langchain.callbacks.manager import CallbackManagerForLLMRun
from langchain.llms.base import LLM

class Guanaco(LLM):
    """Wrapper around AI21 large language models.

    To use, you should have the environment variable ``AI21_API_KEY``
    set with your API key.

    Example:
        .. code-block:: python
    """
    model: str = "baichuan"
    """Model name to use."""

    temperature: float = 0.7
    """What sampling temperature to use."""

    maxTokens: int = 256
    """The maximum number of tokens to generate in the completion."""

    minTokens: int = 0
    """The minimum number of tokens to generate in the completion."""

    topP: float = 1.0
    """Total probability mass of tokens to consider at each step."""

    base_url: Optional[str] = 'http://ecsb.crcloud.com/ecsb/gw/sys/rf?ssdp=QXBpX0lEPWNybGFuZC5jcmxhcnNzLlNDUjAxLjAwMDEmQXBpX1ZlcnNpb249MS4wJkFwcF9TdWJfSUQ9MDAwNTAwMUkwMDAyJkFwcF9Ub2tlbj04M2YwMDE3OTY3NjE0MGRmOWZiNDA0Y2MwNGM3YWRkNyZQYXJ0bmVyX0lEPTAwMDUwMDAwJlNpZ249MjA5MzQzODJkYThkNGJjNGE2OWJhZmQ2MGQxYTJmNzUmU3lzX0lEPTAwMDUwMDFJJlRpbWVfU3RhbXA9MjAyMy0wNS0yMiAxNjowMjoxMjowNjgmVXNlcl9Ub2tlbj0='

    """Base url to use, if None decides based on model name."""

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.forbid


    @property
    def _default_params(self) -> Dict[str, Any]:
        """Get the default parameters for calling AI21 API."""
        return {
            "temperature": self.temperature,
            "maxTokens": self.maxTokens,
            "minTokens": self.minTokens,
            "topP": self.topP,
        }

    @property
    def _identifying_params(self) -> Dict[str, Any]:
        """Get the identifying parameters."""
        return {**{"model": self.model}, **self._default_params}

    @property
    def _llm_type(self) -> str:
        """Return type of llm."""
        return "guanaco"

    def _call(
        self,
        prompt: str,
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
    ) -> str:
        """Call out to AI21's complete endpoint.

        Args:
            prompt: The prompt to pass into the model.
            stop: Optional list of stop words to use when generating.

        Returns:
            The string generated by the model.
        """

        response = requests.post(
            url= self.base_url,
            json={
                "from_server": "646c52e724c3203f886e4bf8a22eddb499e03972", 
                "model": self.model, 
                "message": prompt, 
                "raw_text": "1"
            }
        )
        if response.status_code != 200:
            optional_detail = response.json().get("error")
            raise ValueError(
                f"complete call failed with status code {response.status_code}."
                f" Details: {optional_detail}"
            )
        try:
            response_json = response.json()
            #print(response_json)
            return response_json["data"]
        except Exception as e:
            return e



import random
import json
from typing import Any, Dict, List
from crland.searchhub.page_manage.base import BaseElement
from crland.agent.role import roles_prompts

class Role(BaseElement):
    stype: str = "Role"

class RoleManager:
    type_to_ele: Dict[str, Role] = {}

    def register(
        self, 
        name: str,
        base: Role
    ) -> None:
        if name not in self.type_to_ele:
            self.type_to_ele[name] = base

    def get(
        self, 
        name: str
    ) -> Role:
        return self.type_to_ele.get(name, None)

    # will be moved to db later
    def get_details(
        self, 
        role: str
    ) -> Dict[str, Any]:
        return {
            "role": role,
            "icon": "",
            "id": "",
            "desc": "大佬",
        }

    def get_homepage(
        self, 
        num: int
    ) -> Dict[str, Any]:
        result = []
        for d in roles_prompts.keys():
            result.append(self.get_details(d))
        print(result)
        return result


# flake8: noqa
from typing import Any, Dict, List, Optional, Callable
from crland.searchhub.recall.base import IndexManager
from crland.searchhub.recall.parser import FuncParser
from crland.agent.rank.base import FunctionChain
from crland.datacenter.knowledge.executor import ExecutorManager

from langchain.schema import HumanMessage
from langchain.chains.base import Chain
from langchain.agents.tools import Tool
from langchain.base_language import BaseLanguageModel
from langchain.callbacks.manager import (
    CallbackManagerForChainRun,
)

class FuncRecallChain(Chain):
    llm: BaseLanguageModel
    index_manager: IndexManager
    input_key: str = "input"
    channel_key: str = "channel"
    output_key: str = "func"

    @property
    def input_keys(self) -> List[str]:
        """Will be whatever keys the prompt expects.

        :meta private:
        """
        return [self.input_key]

    @property
    def output_keys(self) -> List[str]:
        """Will always return text key.

        :meta private:
        """
        return [self.output_key]

    @classmethod
    def from_llm_index(
        cls,
        llm : BaseLanguageModel,
        index_manager : IndexManager, 
        **kwargs: Any,  
    ):
        # pass
        return cls(llm=llm, index_manager=index_manager, **kwargs)

    def _call(self,
        inputs: Dict[str, str],
        run_manager: Optional[CallbackManagerForChainRun] = None,
    ) -> Dict[str, Any]:
        
        #docs = self.index_manager.search(
        #    inputs[self.channel_key],
        #    inputs[self.input_key]
        #    )
        docs = list(ExecutorManager().name_to_executor.values())

        tools = []
        tool_names = []
        for exe in docs:
            #exe = id.metadata['ref']
            print("The functions is :", exe.func)
            if exe.func in tool_names:
                continue
            tool_names.append(exe.func)
            tools.append(Tool(
                name = exe.func.__name__,
                func = exe.func,
                return_direct = True,
                description = exe.description
            ))
            # to add args to each function
        AnswerAgent = FunctionChain.from_chains(self.llm, tools)
        #print("AnswerAgent: ", inputs)        
        result = AnswerAgent(inputs)['output']
        print(result)
        prompt = "you are a business analyst, Infer helpfully\n"
        post = "\n Answer:\n"
        if isinstance(result, dict):
            #result['result'] = self.llm(
            #    [HumanMessage(content= prompt + result['result'])]
            #).content
            print(prompt + result['result'] + post)
            result['result'] = self.llm(prompt + result['result'])
        else:
            #result = self.llm(
            #    [HumanMessage(content= prompt + result)]
            #).content
            print(prompt + result + post)
            result = self.llm(prompt + result)
        #print("result", result)
        return {self.output_key: result}


import json
from typing import Any, Dict, List
from crland.tools.embeddings.sentence_transformers import RemoteEmbeddings
from crland.tools.embeddings.huggingface_embedding import Huggingface_embedding
from crland.tools.embeddings.openai_embedding import Openai_embedding

class EmbeddingManager:
    @classmethod
    def get_embedding(self, config):
        try:
            if RemoteEmbeddings.__name__ == config.get("embedding"):
                return RemoteEmbeddings()
            if Huggingface_embedding.__name__ == config.get("embedding"):
                return Huggingface_embedding()
            if Openai_embedding.__name__ == config.get("embedding"):
                return Openai_embedding(config)                
        except:
            pass    
        return RemoteEmbeddings()


from typing import Dict
from abc import ABC, abstractmethod

class Task(ABC):
    name = "task"

    @abstractmethod
    def execute(
        self, 
        text, 
        **kwargs
    ) -> Dict[str, str]:
        pass


import json
from typing import Any, Dict, List, Optional, Callable, Union
from crland.datacenter.knowledge.factor import Factor
from crland.datacenter.knowledge.executor import Executor, ExecutorManager

class Graph:
    name = "Temporal Financial Graph"
    factors: Dict[str, Factor] = {}
    headers: List[str] = []

    def __init__(
        self, 
        factors, 
        headers
    ):
        self.factors = factors
        self.headers = headers

    @classmethod
    def build_from_file(
        cls, 
        filename="crland/datacenter/knowledge/schema.md"
        #infofile="crland/datacenter/knowledge/schema.json",
    ) -> 'Graph':
        '''
            Use Stack to build graph
        '''
        stack = []
        factors = {}
        headers = []
        level = -1
        #name_to_info = json.load(open(infofile, "r"))
        with open(filename, "r") as infile:
            for l in infile:
                data = l.rstrip("").strip("\n").split("-")
                if len(data) == 1:
                    continue
                new_level = int(len(data[0])/3)
                name = data[1].strip()
                if name not in factors:
                    factors.update(
                        {
                            name: Factor(
                                name, 
                                "",
                                [],
                                [],
                                [],
                                [])
                        }  # a deep hole here for python object memory allocation
                    )
                if 0 == new_level:
                    headers.append(name)      
                if new_level > level:
                    stack.append(name)
                elif new_level == level:
                    stack[-1] = name
                else:
                    stack = stack[:new_level]
                    stack.append(name)
                level = new_level
                if level > 0:
                    factors[name].add_parents(
                        factors[stack[-2]]
                    )
                    factors[stack[-2]].add_childrens(
                        factors[name]
                    )
        return cls(factors, headers)

    def show(
        self
    ):
        for n, f in self.factors.items():
            print("*"*10, f.name, "*"*10)
            for p in f.get_parents():
                print("parents:", p.name)
            for c in f.get_childrens():
                print("children:", c.name)
    
    def get_factor(
        self, 
        name
    ) -> Factor:
        if name in self.factors:
            return self.factors[name]
        raise f"No Element called {name}"

    #  downgrade, will drop in future
    def assemble_func(
        self, 
        name, 
        func
    ):
        self.factors[name].register_func(
            Executor.create_from_func(func, self.factors[name].description)
        )

    #  downgrade, will drop in future
    def assemble(
        self, 
        funcs: Union[List[Dict[str, Callable]], Dict[str, Callable]]
    ):
        if isinstance(funcs, List):
            for term in funcs:
                for name, func in term.items():
                    self.assemble_func(name, func)
        else:
            for name, func in funcs.items():
                self.assemble_func(name, func)

    def assemble(
        self, 
        manager: ExecutorManager
    ):
        for name_signature, exe in ExecutorManager().name_to_executor.items():
            d = name_signature.split("|")
            name = d[0]
            signature = d[1]
            if not exe.graph_node:
                continue
            if signature == "default":
                self.factors[name].description = exe.description
            if name in self.factors:
                self.factors[name].register_func(
                    exe
                )

    def get_available_executors(
        self,
        factor_to_models: Dict[str, str] = {}
    ):
        '''
            Based on user's requirements, return funcs
            :factor_to_models, user require
        '''
        desc_to_func = {}
        for name, factor in self.factors.items():
            if len(factor.executors):
                flag = False
                model = factor_to_models.get(name, "")
                if model:  # if user require hit
                    for exe in factor.executors:
                        if model == exe.name:
                            desc_to_func.update({factor.name : exe})  # recall by name
                            desc_to_func.update({factor.description : exe})  # recall by description                                                     
                            flag = True
                            break
                if not flag:
                    desc_to_func.update({factor.name : factor.executors[0]})
                    desc_to_func.update({factor.description : factor.executors[0]})
        return desc_to_func

    def get_all_exec(
        self
    ):
        '''
            Based on user's requirements, return execs
        '''
        desc_to_func = []
        for name, factor in self.factors.items():
            for exe in factor.executors:
                desc_to_func.append((name, exe))
        return desc_to_func

    def get_factor_exec(
        self,
        factor: str
    ):
        '''
            return factor and children func
        '''
        desc_to_func = []
        factor = self.factors[factor]
        def inner_func(factor, desc_to_func):
            if len(factor.executors):
                for exe in factor.executors:
                    desc_to_func.append((factor.name, exe))
            for chd in factor.get_childrens():
                desc_to_func = inner_func(chd, desc_to_func)
            return desc_to_func
        return inner_func(factor, desc_to_func)


if __name__ == "__main__":
    graph = Graph.build_from_file("crland/datacenter/knowledge/schema.md")
    graph.show()


import json

class Config:
    def __init__(
        self, 
        filename="crland/config/config.json"
    ):
        with open(filename, 'r') as f:
            self.config = json.load(f)
    
    def get(
        self, 
        key
    ):
        return self.config.get(key)
    
    def set(
        self, 
        key, 
        value
    ):
        self.config[key] = value

    def save(
        self, 
        filename
    ):
        with open(filename, 'w') as f:
            json.dump(self.config, f)


# flake8: noqa
from langchain.prompts import PromptTemplate

role_prompt_template = """you are Warren Buffett, you must consider quesiton based on the following factors
```
1. Focus on the business fundamentals
2. Buy at a discount
3. Hold for the long run
4. Focus on cash flow and earnings
5. Focus on simple, stable companies
6. Focus on management teams
7. Margin of safety
8. Hold a concentrated portfolio
```
you must answer question in Warren Buffett's tone and think based on Warren Buffett's mind.

begin to answer! Question is {input}.

Answer:
"""
Warren_Buffett_PROMPT = PromptTemplate(
    template=role_prompt_template, input_variables=["input"])


import json
from typing import Any, Dict, List
from crland.searchhub.page_manage.base import BaseElement
from crland.datacenter.database.external.qstock.trade import stock_billboard
from crland.datacenter.database.base import Database

class Company(BaseElement):
    stype: str = "Company"

class CompanyManager:

    def __init__(
        self,
        config
    ):
        self.db = Database(config.get("db"))
        self.type_to_ele: Dict[str, Company] = {}

    def register(
        self, 
        name: str, 
        base: Company
    ) -> None:
        if name not in self.type_to_ele:
            self.type_to_ele[name] = base

    def get(
        self, 
        name: str
    ) -> Company:
        return self.type_to_ele.get(name, None)

    # will be moved to db later
    def get_details(
        self, 
        company: str
    ) -> Dict[str, Any]:
        return {
            "company": company,
            "icon": "",
            "id": "",
        }

    def get_homepage(
        self, 
        num: int
    ) -> Dict[str, Any]:
        result = []
        data = stock_billboard()['股票名称'].to_list()
        for d in data:
            result.append(self.get_details(d))
        #print(result)
        return result

    def search(
        self,
        text: str,
        num: int
    ) -> Dict[str, Any]:
        result = []
        data = self.db.select_more(
            table = "t_balance_sheet_statement",
            range_str = f"SECURITY_NAME like '%{text}%'",
            field = "distinct(SECURITY_NAME)",
        )
        #print("data", data)
        for d in data:
            result.append(self.get_details(d['SECURITY_NAME']))
        return result




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="crland",
    version="0.2",
    author="Bin ZHU",
    author_email="",
    description="llm agent framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="",
    packages=setuptools.find_packages(),
    install_requires=['langchain==0.0.266', 'py_mini_racer==0.6.0'],
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ),
)


from typing import Any, Dict, List, Optional, Callable

class ChartManager:
    name = "ChartManager"
    name_to_func: Dict[str, Any] = {}

    def register(
        self,
        name,
        func
    ):
        if name not in self.name_to_func:
            self.name_to_func[name] = func
    
    def get(
        self,
        name
    ):
        if name in self.name_to_func:
            return self.name_to_func[name]
        else:
            return None


from typing import Union
from langchain.schema import HumanMessage
from langchain.chat_models.base import BaseChatModel
from langchain.llms.base import BaseLLM
# a quick func to test

class JustModel:
    def __init__(self, llm):
        self.llm = llm
    def __call__(self, text):
        if isinstance(self.llm, BaseChatModel):
            return {"result": self.llm([HumanMessage(content=text)]).content}
        elif isinstance(llm, BaseLLM):
            return {"result": self.llm(text)}


from functools import wraps
from crland.datacenter.knowledge.executor import ExecutorManager

def register(name, description, signature="default", graph_node=True):
    #print("register")
    def call(func):
        #print("call")
        ExecutorManager().register(
            name = name,
            func = func,
            description = description,
            signature = signature,
            graph_node = graph_node
        )
        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)
        return wrapper       
    return call


from typing import Any, List, Optional, Sequence, Tuple, Dict, Union

from pydantic import Field

from langchain.agents.agent import Agent, AgentExecutor, AgentOutputParser

from crland.agent.rank.prompt import FORMAT_INSTRUCTIONS, PREFIX, SUFFIX
from crland.agent.rank.output_parser import FunctionOutParser

from langchain.base_language import BaseLanguageModel
from langchain.callbacks.base import BaseCallbackManager
from langchain.chains.llm import LLMChain
from langchain.prompts.prompt import PromptTemplate
from langchain.schema import AgentAction
from langchain.agents.tools import Tool
from langchain.tools.base import BaseTool
from langchain.agents.utils import validate_tools_single_input
from langchain.callbacks.manager import (
    CallbackManagerForChainRun,
)
from langchain.schema import (
    AgentAction,
    AgentFinish,
)
from langchain.input import get_color_mapping

class FunctionAgent(Agent):
    """Agent for the FunctionAgent chain."""

    output_parser: AgentOutputParser = Field(default_factory=FunctionOutParser)

    @classmethod
    def _get_default_output_parser(cls, **kwargs: Any) -> AgentOutputParser:
        return FunctionOutParser()

    @property
    def _agent_type(self) -> str:
        """Return Identifier of agent type."""
        return "FunctionAgent"

    @property
    def observation_prefix(self) -> str:
        """Prefix to append the observation with."""
        return "Observation: "

    @property
    def llm_prefix(self) -> str:
        """Prefix to append the llm call with."""
        return "Thought:"

    @classmethod
    def create_prompt(
        cls,
        tools: Sequence[BaseTool],
        prefix: str = PREFIX,
        suffix: str = SUFFIX,
        format_instructions: str = FORMAT_INSTRUCTIONS,
        input_variables: Optional[List[str]] = None,
    ) -> PromptTemplate:
        """Create prompt in the style of the zero shot agent.

        Args:
            tools: List of tools the agent will have access to, used to format the
                prompt.
            prefix: String to put before the list of tools.
            suffix: String to put after the list of tools.
            input_variables: List of input variables the final prompt will expect.

        Returns:
            A PromptTemplate with the template assembled from the pieces here.
        """
        tool_strings = "\n".join([f"{tool.name}: {tool.description}" for tool in tools])
        tool_names = ", ".join([tool.name for tool in tools])
        format_instructions = format_instructions.format(tool_names=tool_names) + """: {query}"""
        template = "\n\n".join([prefix, tool_strings, format_instructions, suffix])
        #print(template)
        if input_variables is None:
            input_variables = ["input", "agent_scratchpad", "query"]
        return PromptTemplate(template=template, input_variables=input_variables)

    @classmethod
    def from_llm_and_tools(
        cls,
        llm: BaseLanguageModel,
        tools: Sequence[BaseTool],
        callback_manager: Optional[BaseCallbackManager] = None,
        output_parser: Optional[AgentOutputParser] = None,
        prefix: str = PREFIX,
        suffix: str = SUFFIX,
        format_instructions: str = FORMAT_INSTRUCTIONS,
        input_variables: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> Agent:
        """Construct an agent from an LLM and tools."""
        cls._validate_tools(tools)
        prompt = cls.create_prompt(
            tools,
            prefix=prefix,
            suffix=suffix,
            format_instructions=format_instructions,
            input_variables=input_variables,
        )
        llm_chain = LLMChain(
            llm=llm,
            prompt=prompt,
            callback_manager=callback_manager,
        )
        tool_names = [tool.name for tool in tools]
        _output_parser = output_parser or cls._get_default_output_parser()
        return cls(
            llm_chain=llm_chain,
            allowed_tools=tool_names,
            output_parser=_output_parser,
            **kwargs,
        )

    @classmethod
    def _validate_tools(cls, tools: Sequence[BaseTool]) -> None:
        validate_tools_single_input(cls.__name__, tools)
        for tool in tools:
            if tool.description is None:
                raise ValueError(
                    f"Got a tool {tool.name} without a description. For this agent, "
                    f"a description must always be provided."
                )
        super()._validate_tools(tools)

class FunctionChain(AgentExecutor):
    """Chain that implements the MRKL system.
    """
    @classmethod
    def from_chains(
        cls, llm: BaseLanguageModel, tools: List[Tool], **kwargs: Any
    ) -> AgentExecutor:
        """User friendly way to initialize the MRKL chain.

        This is intended to be an easy way to get up and running with the
        MRKL chain.

        Args:
            llm: The LLM to use as the agent LLM.
            tools: The chains the MRKL system has access to.
            **kwargs: parameters to be passed to initialization.

        Returns:
            An initialized MRKL chain.

        """
        agent = FunctionAgent.from_llm_and_tools(llm, tools)
        return cls(agent=agent, tools=tools, **kwargs)
        
    def _call(
        self,
        inputs: Dict[str, str],
        run_manager: Optional[CallbackManagerForChainRun] = None,
    ) -> Dict[str, Any]:
        """Run text through and get agent response."""
        # Construct a mapping of tool name to tool for easy lookup
        name_to_tool_map = {tool.name: tool for tool in self.tools}
        # We construct a mapping from each tool to a color, used for logging.
        color_mapping = get_color_mapping(
            [tool.name for tool in self.tools], excluded_colors=["green"]
        )
        intermediate_steps: List[Tuple[AgentAction, str]] = []
        # We now enter the agent loop (until it returns something).
        next_step_output = self._take_next_step(
            name_to_tool_map,
            color_mapping,
            inputs,
            intermediate_steps,
            run_manager=run_manager,
        )
        if isinstance(next_step_output, AgentFinish):
            return self._return(
                next_step_output, intermediate_steps, run_manager=run_manager
            )
        intermediate_steps.extend(next_step_output)
        if len(next_step_output) == 1:
            next_step_action = next_step_output[0]
            # See if tool should return directly
            # print("next_step_action", next_step_action)
            tool_return = self._get_tool_return(next_step_action)
            #print("tool_return", next_step_action)
            if tool_return is not None:
                return self._return(tool_return, intermediate_steps)
        return {"output": "Sorry! Not found"}

    def _take_next_step(
        self,
        name_to_tool_map: Dict[str, BaseTool],
        color_mapping: Dict[str, str],
        inputs: Dict[str, str],
        intermediate_steps: List[Tuple[AgentAction, str]],
        run_manager: Optional[CallbackManagerForChainRun] = None,
    ) -> Union[AgentFinish, List[Tuple[AgentAction, str]]]:
        """Take a single step in the thought-action-observation loop.

        Override this to take control of how the agent makes and acts on choices.
        """
        try:
            # Call the LLM to see what to do.
            output = self.agent.plan(
                intermediate_steps,
                callbacks=run_manager.get_child() if run_manager else None,
                **inputs,
            )
        except Exception as e:
            if not self.handle_parsing_errors:
                raise e
            text = str(e).split("`")[1]
            observation = "Invalid or incomplete response"
            output = AgentAction("_Exception", observation, text)
            tool_run_kwargs = self.agent.tool_run_logging_kwargs()
            observation = ExceptionTool().run(
                output.tool,
                verbose=self.verbose,
                color=None,
                callbacks=run_manager.get_child() if run_manager else None,
                **tool_run_kwargs,
            )
            return [(output, observation)]
        # If the tool chosen is the finishing tool, then we end and return.
        if isinstance(output, AgentFinish):
            return output
        actions: List[AgentAction]
        if isinstance(output, AgentAction):
            actions = [output]
        else:
            actions = output
        result = []
        for agent_action in actions:
            # if have input params, use default
            if "params" in inputs and len(inputs["params"]) != 0:
                #print(inputs['params'])
                agent_action.tool_input=inputs['params']
            #if run_manager:
            #    run_manager.on_agent_action(agent_action, color="green")
            # Otherwise we lookup the tool
            #print(agent_action.tool, name_to_tool_map)
            if agent_action.tool in name_to_tool_map:
                tool = name_to_tool_map[agent_action.tool]
                return_direct = tool.return_direct
                color = color_mapping[agent_action.tool]
                tool_run_kwargs = self.agent.tool_run_logging_kwargs()
                if return_direct:
                    tool_run_kwargs["llm_prefix"] = ""
                # We then call the tool on the tool input to get an observation
                # if inputs are multi-variable
                if isinstance(agent_action.tool_input, list):
                    observation = ""
                    for k in agent_action.tool_input:
                        observation += k + ":\n"
                        observation += tool.run(
                            k,
                            verbose=self.verbose,
                            color=color,
                            callbacks=run_manager.get_child() if run_manager else None,
                            **tool_run_kwargs,
                        )
                        observation += "\n"
                else:
                    observation = tool.run(
                        agent_action.tool_input,
                        verbose=self.verbose,
                        color=color,
                        callbacks=run_manager.get_child() if run_manager else None,
                        **tool_run_kwargs,
                    )
            else:
                tool_run_kwargs = self.agent.tool_run_logging_kwargs()
                observation = InvalidTool().run(
                    agent_action.tool,
                    verbose=self.verbose,
                    color=None,
                    callbacks=run_manager.get_child() if run_manager else None,
                    **tool_run_kwargs,
                )
            result.append((agent_action, observation))
        return result


from typing import Any, Awaitable, Callable, Dict, Optional, Tuple, Type, Union
from crland.datacenter.knowledge.factor import Factor

EMPTY_DATA = "No Data"
EMPTY_NUM = -1000000

def factor(
    func: Callable,
    factor: Factor,
    return_direct: bool = False,
    ) -> Callable:
    """Make tools out of functions, can be used with or without arguments.

    Args:
        *args: The arguments to the tool.
        return_direct: Whether to return directly from the tool rather
            than continuing the agent loop.

    Requires:
        - Function must be of type (str) -> str
        - Function must have a docstring

    Examples:
        .. code-block:: python
            @factor(factor= your Factor, return_direct=True)
            def search_api(query: str) -> str:
                # Searches the API for the query.
                return
    """
    def _make_tool(func: Callable) -> BaseTool:
        assert func.__doc__ is not None, "Function must have a docstring"
        tool = Tool(
            name=func.__name__,
            func=func,
            description=f"tool",
            return_direct=return_direct,
        )
        factor.register(tool)
        return tool
    return _make_tool(func)